/*
 * Copyright (C) 2004, 2005 Moritz Orbach <zufall@apfelboymchen.homeunix.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <time.h>
#include <signal.h>

#include <X11/Xutil.h>
#include <Imlib2.h>

#include "config.h"

#include "list.h"
#include "setroot.h"
#include "cfg.h"
#include "status.h"
#include "recurse.h"

#ifdef FENCE
#include "efence.h"
#endif

/* vielleicht
 * - vernnftige hintergrundfarbe:
 *   . zufall ohne grelle farben
 *   . r, g und b max- bzw. minimalwerte festlegen
 * - datei auslesen in denen ausnahmen zu default optionen stehen (pro bild). ugh.
 */

t_list *imglist;
Display *Xdisplay = NULL;
Screen *scr;
Window Xroot;
int screen;

unsigned int getrandom(unsigned int max)
{
	if (max > 0)
		return(rand() % max + 1);
	else
		return 0;
} 

void exit_func(void)
{
	extern t_list *imglist;

	if (!list_destroy(imglist))
		fprintf(stderr, "error freeing memory..\n");
	if (Xdisplay) XCloseDisplay(Xdisplay);
}

void initX(void)
{
	/* clients ohne windows tauchen anscheinend nicht in xlsclients auf */
	Xdisplay = XOpenDisplay(NULL);
	if (!Xdisplay) {
		fprintf(stderr, "%s: Unable to open display\n", PACKAGE_NAME);
		exit(1);
	}

	if (atexit(exit_func)) {
		fprintf(stderr,"zufall: atexit() failed\n");
		exit(2);
	}

	screen = DefaultScreen(Xdisplay);
	Xroot = RootWindow(Xdisplay, screen);
	scr = ScreenOfDisplay(Xdisplay, screen);
	imlib_context_set_display(Xdisplay);
	imlib_context_set_visual(DefaultVisual(Xdisplay, DefaultScreen(Xdisplay)));
	imlib_context_set_anti_alias(1);
	imlib_context_set_dither(1);
	imlib_context_set_blend(1);
}

/* signal handlers */

void infotrap(p)
{
	unsigned int tmp;
	tmp = list_allocated();

	/*
	 * "calling library functions from a signal handler is not guaranteed to work 
	 * in all implementations....."
	 * http://publications.gbdirect.co.uk/c_book/chapter9/signal_handling.html
	 * http://www.delorie.com/gnu/docs/glibc/libc_488.html
	 * hm..
	 */
	/* warning: ISO C90 does not support the `%lf' printf format */
	printf("allocated: %d b (%.2lf kb), %d nodes\n", tmp, (double)tmp / 1024, list_count(imglist));
	signal(SIGUSR1, infotrap);
	break_sleep(0);
}

void nop(p)
{
	/* NOP */
	signal(SIGHUP, nop);
	break_sleep(1);
}

void signal_quit()
{
	printf("%s: exiting\n", PACKAGE_NAME);
	halt();
	break_sleep(1);
}

/* --------- */

int main(int argc, char *argv[])
{
	extern t_list *imglist;
	char *image;
	unsigned int zufall;
	unsigned int tr;	/* time remaining */
	int error = 0;		/* return code */
	int diroffset = 0;	/* grr :( */

	/* killall / CTRL-C */
	signal(SIGTERM, signal_quit);
	signal(SIGINT, signal_quit);

	/*
	 * don't buffer output more than a line, so that the filename can be 
	 * seen immediately with tail -f when the image changes.
	 * "The standard error stream stderr is always unbuffered by default."
	 */
	setvbuf(stdout, (char *)NULL, _IOLBF, 0);

	set_defaults();
	/* cfg_file(); */
	cfg_cmdln(&argc, argv, &diroffset);
	argc = argc - diroffset;
	argv = &argv[diroffset];

	if (running() && !error) {
		initX();
		imglist = list_init();
		srand(time (0));

		buildlist(imglist, argc, argv);
	}

	if (list_count(imglist) > 0) {
		/* imagechange, info */
		signal(SIGHUP, nop);
		signal(SIGUSR1, infotrap);
	} else {
		error++;
		halt();
		fprintf(stderr, "Sorry, no supported images found\n");
	}

	while (running() && !error) {

		zufall = getrandom(list_count(imglist));
		image = list_get(imglist, zufall);
		if (image == NULL) {
			/* more than 0 images, but there is no filename?? */
			halt();
			fprintf(stderr, "internal error. pardon!\n");
			break;
		}
		printf("so ein zufall!: %s\n", image);

		/* don't sleep if there is only one image left */
		if (list_count(imglist) > 1)
			tr = cfg.delay;
		else {
			tr = 0;
			printf("no pictures left, exiting\n");
			halt();
		}
		if (setroot(Xdisplay, Xroot, scr, screen, image, "")) {
			while (tr != 0) {
				/* hintergrund bewundern */
				tr = sleep(tr);
				/* weiterschlafen oder neu schlafen? */
				if (sleep_broken()) {
					/* reset fr bildwechsel oder ende */
					tr = 0;
					break_sleep(0);
				} else {
					/* bildwechsel / signal */
					if (tr != 0) {
						printf("%s\n", image);
						printf("%d seconds (%.1d:%.2d) remaining\n", tr, (int)tr / 60, (int)tr % 60);
					}
				}
					
			}
		} else {
			fprintf(stderr, "Unable to load image file \"%s\".\n", image);
			if (!list_delete(imglist, zufall)) {
				fprintf(stderr, "error freeing memory..\n");
				error++;
			}
		}
	}

	exit(error);
}
